class BanhammerException(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return str(self.msg)


class NotModerator(BanhammerException):
    def __init__(self, user, sub):
        super(NotModerator, self).__init__("/u/{} doesn't moderate /r/{}.".format(user, sub))


class NoRedditInstance(BanhammerException):
    def __init__(self):
        super(NoRedditInstance, self).__init__(
            "No <apraw.Reddit> instance was given to the <banhammer.Subreddit> object.")


class NoItemGiven(BanhammerException):
    def __init__(self):
        super(NoItemGiven, self).__init__("No <banhammer.RedditItem> was given to the <banhammer.Reaction> instance.")


class NotEligibleItem(BanhammerException):
    def __init__(self):
        super(NotEligibleItem, self).__init__(
            "The <banhammer.RedditItem> given to the <banhammer.Reaction> object cannot be handled.")



from apraw.models import (Comment, ModmailConversation, ModmailMessage,
                          Submission)

from ..exceptions import NoItemGiven, NotEligibleItem
from ..utils import yaml
from .item import RedditItem


class ReactionPayload:

    def __init__(self, user="Banhammer"):
        self.item = None
        self.user = user
        self.actions = list()
        self.approved = False
        self.reply = ""
        self.emoji = ""

    def feed(self, item, approved, user="", emoji="", reply=""):
        self.item = item
        self.approved = approved
        self.user = user or self.user
        self.emoji = emoji
        self.reply = reply

    async def get_message(self):
        if len(self.actions) == 0:
            self.actions.append("dismissed")
        return f"**{self.item.type.title()} {' and '.join(self.actions)} by {self.user}!**\n\n" \
               f"{self.item.type.title()} by /u/{await self.item.get_author_name()}:\n\n" \
               f"{self.item.url}"


class ReactionHandler:

    async def handle(self, reaction, item, payload):
        if isinstance(item.item, (ModmailConversation, ModmailMessage)):
            conversation = item.item.conversation if isinstance(item, ModmailMessage) else item.item
            if reaction.archive:
                await conversation.archive()
                payload.actions.append("archived")
            if reaction.mute:
                await conversation.mute()
                payload.actions.append("muted")
            if reaction.reply != "":
                await conversation.reply(reaction.reply)
                payload.actions.append("replied to")
            return payload

        if await item.is_author_removed():
            await item.item.mod.remove()
            payload.actions.append("removed")
            await item.item.mod.lock()
            payload.actions.append("locked")

            payload.feed(item, False, "Banhammer")
            return payload

        if reaction.approve:
            await item.item.mod.approve()
            payload.actions.append("approved")
        else:
            await item.item.mod.remove()
            payload.actions.append("removed")

        if reaction.lock or not reaction.approve:
            await item.item.mod.lock()
            payload.actions.append("locked")
        else:
            if item.item.locked:
                await item.item.mod.unlock()
                payload.actions.append("unlocked")

        if isinstance(item.item, Submission):
            if reaction.flair:
                await item.item.mod.flair(text=reaction.flair)
                payload.actions.append("flaired")

            if reaction.mark_nsfw:
                await item.item.mod.nsfw()
                payload.actions.append("marked NSFW")

        if reaction.reply:
            reply = await item.item.reply(reaction.reply)
            if reaction.distinguish_reply:
                await reply.mod.distinguish(sticky=reaction.sticky_reply)
            payload.actions.append("replied to")

        if isinstance(reaction.ban, int):
            ban_message = item.subreddit.banhammer.message_builder.get_ban_message(item, reaction.ban)
            if reaction.ban == 0:
                subreddit = await item.item.subreddit()
                await subreddit.banned.add(item.item.author.name, ban_reason="Breaking Rules",
                                           ban_message=ban_message, note="Banhammer Ban")
                payload.actions.append("/u/" + item.item.author.name + " permanently banned")
            else:
                subreddit = await item.item.subreddit()
                await subreddit.banned.add(item.item.author.name, ban_reason="Breaking Rules",
                                           duration=reaction.ban, ban_message=ban_message,
                                           note="Banhammer Ban")
                payload.actions.append(f"/u/{item.item.author.name} banned for {reaction.ban} day(s)")

        return payload


class Reaction:

    def __init__(self, **kwargs):
        self.config = kwargs

        self.emoji = kwargs.get("emoji", "").strip()

        self.type = kwargs.get("type", "")
        self.flair = kwargs.get("flair", "")
        self.approve = kwargs.get("approve", False)
        self.mark_nsfw = kwargs.get("mark_nsfw", False)
        self.lock = kwargs.get("lock", False)
        self.reply = kwargs.get("reply", "")

        self.sticky_reply = kwargs.get("sticky_reply", True)
        self.distinguish_reply = kwargs.get("distinguish_reply", True) or self.sticky_reply

        self.ban = kwargs.get("ban", None)
        self.archive = kwargs.get("archive", False)
        self.mute = kwargs.get("mute", False)
        self.min_votes = kwargs.get("min_votes", 1)

    def __str__(self):
        str = self.emoji

        if self.type in ["submission", "comment", ""]:
            if self.type:
                str += " | " + self.type
            else:
                str += " | submissions + comments"
            if self.flair:
                str += " | flair: " + self.flair
            str += " | " + ("approve" if self.approve else "remove")
            if self.mark_nsfw:
                str += " | mark NSFW"
            if self.lock or not self.approve:
                str += " | lock"
            if self.ban is not None:
                str += " | " + ("permanent ban" if self.ban == 0 else f"{self.ban} day ban")
        if self.reply:
            str += " | reply"
        if self.min_votes:
            str += f" | min votes: {self.min_votes}"

        return str

    async def handle(self, item, payload=ReactionPayload(), user=""):
        if not self.eligible(item.item):
            raise NotEligibleItem()

        payload.feed(item, self.approve, user or payload.user, self.emoji, self.reply)

        return await item.subreddit.banhammer.reaction_handler.handle(self, item, payload)

    def eligible(self, item):
        if isinstance(item, Submission):
            if self.type == "" or self.type == "submission":
                return True
        elif isinstance(item, Comment):
            if self.type == "" or self.type == "comment":
                return True
        elif isinstance(item, (ModmailMessage, ModmailConversation)):
            if self.type == "mail":
                return True
        return False


def get_reactions(yml):
    result = yaml.get_list(yml)
    ignore = list()
    emojis = set()
    for item in result:
        if "ignore" in item:
            ignore = [i.strip() for i in item["ignore"].split(",")]
            result.remove(item)
            break
    reactions = result
    reactions = [Reaction(**r) for r in result if "emoji" in r]
    return {
        "ignore": ignore,
        "reactions": reactions
    }


def ignore_reactions(reactions, remove):
    emojis = set(i.emoji if isinstance(item, Reaction) else i for i in remove)
    reactions = [r for r in reactions if react.emoji not in emojis]
    return reactions



import logging

import apraw
import discord
from apraw.models import (Comment, Message, ModmailConversation,
                          ModmailMessage, Submission, Subreddit)

logger = logging.getLogger("banhammer")


class RedditItem:

    def __init__(self, item, subreddit, source):
        self.item = item
        self.id = item.id
        self.type = "submission" if isinstance(
            item, Submission) else "comment" if isinstance(
            item, Comment) else "modmail" if type(item) in [
            ModmailMessage, ModmailConversation] else "mod action"
        self.subreddit = subreddit
        self.source = source
        self._author = None

    async def get_message(self):
        return await self.subreddit.banhammer.message_builder.get_item_message(self)

    async def get_embed(self, embed_color: discord.Color = None):
        return await self.subreddit.banhammer.message_builder.get_item_embed(self, embed_color)

    async def get_author(self):
        if not self._author:
            if not isinstance(self.item, ModmailConversation):
                try:
                    self._author = await self.item.author()
                except Exception as e:
                    logger.error(e)
            else:
                self._author = self.item.authors[0]
        return self._author

    async def is_author_removed(self):
        author = await self.get_author()
        author_removed = not author
        if author and not isinstance(author, dict):
            author_removed = not hasattr(author, "name")
        elif author:
            author_removed = author.get("isDeleted", False)
        return author_removed

    async def get_author_name(self):
        author = await self.get_author()
        if await self.is_author_removed():
            return "[deleted]"
        else:
            return author.name

    def get_reactions(self):
        return self.subreddit.get_reactions(self.item)

    async def add_reactions(self, message):
        for r in self.get_reactions():
            try:
                await message.add_reaction(r.emoji)
            except Exception as e:
                logger.error(e)
                continue

    def get_reaction(self, emoji):
        return self.subreddit.get_reaction(emoji, self.item)

    @property
    def url(self):
        return get_item_url(self.item)


def get_item_url(item):
    if isinstance(item, Submission):
        return f"https://www.reddit.com/r/{item._data['subreddit']}/comments/{item.id}"
    elif isinstance(item, Comment):
        return f"https://www.reddit.com/r/{item._data['subreddit']}/comments/{item._data['link_id']}/_/{item.id}"
    elif isinstance(item, ModmailConversation):
        return "https://mod.reddit.com/mail/all/" + item.id
    elif isinstance(item, ModmailMessage):
        return "https://mod.reddit.com/mail/all/" + item.conversation.id
    elif isinstance(item, Message):
        if item.was_comment:
            return f"https://www.reddit.com/r/{item._data['subreddit']}/comments/{item._data['link_id']}/_/{item.id}"
        else:
            return "https://www.reddit.com/message/messages/{}" + str(item)
    elif isinstance(item, Subreddit):
        return "https://www.reddit.com/r/" + item.display_name
    return ""



import discord

from ..const import BOT_DISCLAIMER
from .item import RedditItem


class MessageBuilder:
    async def get_item_message(self, item: RedditItem):
        if item.type in ["submission", "comment"]:
            return f"New {item.type} on /r/{item.item._data['subreddit']} by /u/{await item.get_author_name()}!\n\n" + \
                f"https://www.reddit.com{item.item.permalink}\n\n" + \
                f"**Title:** {item.item.title}\n**Body:**\n{item.item.selftext}"
        elif item.type == "modmail":
            return f"New message in modmail conversation '{item.item.conversation.subject}' on /r/{item.item.conversation._data['owner']} by /u/{await item.get_author_name()}!" + \
                f"\n\n{item.item.body_md}"
        else:
            return f"New action taken by /u/{item.item._data['mod']} on /r/{item.item.subreddit}: `{item.item.action}`"

    async def get_item_embed(self, item: RedditItem, embed_color: discord.Color = None):
        embed = discord.Embed(
            colour=embed_color or item.subreddit.banhammer.embed_color
        )

        title = ""
        if item.type in ["submission", "comment"]:
            if item.source == "reports":
                title = f"{item.type.title()} reported on /r/{item.item._data['subreddit']} by /u/{await item.get_author_name()}!"
            else:
                title = f"New {item.type} on /r/{item.item._data['subreddit']} by /u/{await item.get_author_name()}!"
        elif item.type == "modmail":
            title = f"New message in modmail conversation '{item.item.conversation.subject}' on /r/{item.item.conversation._data['owner']} by /u/{await item.get_author_name()}!"
        else:
            title = f"New action taken by /u/{item.item._data['mod']} on /r/{item.item._data['subreddit']}!"

        url = item.url
        embed.set_author(name=title, url=url if url else discord.Embed.Empty)

        if item.type == "submission":
            embed.add_field(name="Title", value=item.item.title, inline=False)
            if item.item.is_self:
                embed.add_field(name="Body",
                                value=item.item.selftext if item.item.selftext else "Empty",
                                inline=False)
            else:
                embed.add_field(name="URL", value=item.item.url, inline=False)
            if item.source == "reports":
                reports = [f"{r[1]} {r[0]}" for r in item.item.user_reports]
                embed.add_field(name="Reports", value="\n".join(reports), inline=False)
        elif item.type == "comment":
            embed.description = item.item.body
        elif item.type == "modmail":
            embed.description = item.item.body_md
        elif item.type == "mod action":
            embed.description = f"Action: `{item.item.action}`"

        return embed

    def get_ban_message(self, item, ban_duration):
        ban_type = "permanent" if not ban_duration else "temporary"
        disclaimer = BOT_DISCLAIMER.format(item.subreddit.get_contact_url())
        return f"Our moderator team has reviewed [this post]({item.url}) and decided to give you a {ban_type} ban. " \
               f"If you wish to appeal this ban, please respond to this message.\n\n{disclaimer}"



import asyncio
import json
import logging
import os
import re

import apraw
import discord
from apraw.utils import ExponentialCounter

from .models import MessageBuilder, ReactionHandler, Subreddit
from .utils import reddit_helper

banhammer_purple = discord.Colour(0).from_rgb(207, 206, 255)
logger = logging.getLogger("banhammer")


class Banhammer:

    def __init__(self, reddit: apraw.Reddit, max_loop_time: int = 16, bot: discord.Client = None, embed_color: discord.Colour = banhammer_purple,
                 change_presence: bool = False, message_builder: MessageBuilder = MessageBuilder(), reaction_handler: ReactionHandler = ReactionHandler()):
        self.reddit = reddit
        self.subreddits = list()
        self.loop = asyncio.get_event_loop()

        self.item_funcs = list()
        self.action_funcs = list()

        self.max_loop_time = max_loop_time

        self.message_builder = message_builder
        self.reaction_handler = reaction_handler
        self.bot = bot
        self.embed_color = embed_color
        self.change_presence = change_presence

    async def add_subreddits(self, *subs):
        for sub in subs:
            if not isinstance(sub, Subreddit):
                sub = Subreddit(self, subreddit=str(sub))
                await sub.load_reactions()
            self.subreddits.append(sub)

    def remove_subreddit(self, subreddit):
        subreddit = str(subreddit).lower().replace("r/", "").replace("/", "")
        for sub in self.subreddits:
            sub = str(sub).lower().replace("r/", "").replace("/", "")
            if sub == subreddit:
                self.subreddits.remove(sub)
                return True
        return False

    def new(self, **kwargs):
        def assign(func):
            self.add_new_func(func, **kwargs)
            return func

        return assign

    def add_new_func(self, func, **kwargs):
        self.add_items_func(func, "get_new", **kwargs)

    def comments(self, **kwargs):
        def assign(func):
            self.add_comments_func(func, **kwargs)
            return func

        return assign

    def add_comments_func(self, func, **kwargs):
        self.add_items_func(func, "get_comments", **kwargs)

    def mail(self, **kwargs):
        def assign(func):
            self.add_mail_func(func, **kwargs)
            return func

        return assign

    def add_mail_func(self, func, **kwargs):
        self.add_items_func(func, "get_mail", **kwargs)

    def queue(self, **kwargs):
        def assign(func):
            self.add_queue_func(func, **kwargs)
            return func

        return assign

    def add_queue_func(self, func, **kwargs):
        self.add_items_func(func, "get_queue", **kwargs)

    def reports(self, **kwargs):
        def assign(func):
            self.add_report_func(func, **kwargs)
            return func

        return assign

    def add_report_func(self, func, **kwargs):
        self.add_items_func(func, "get_reports", **kwargs)

    def add_items_func(self, func, sub_func, **kwargs):
        if asyncio.iscoroutinefunction(func):
            self.item_funcs.append({
                "func": func,
                "sub": kwargs["subreddit"] if "subreddit" in kwargs else None,
                "sub_func": sub_func
            })

    async def send_items(self):
        counter = ExponentialCounter(self.max_loop_time)

        while True:
            found = False

            if self.bot and self.change_presence:
                try:
                    watching = discord.Activity(type=discord.ActivityType.watching, name="Reddit")
                    await self.bot.change_presence(activity=watching)
                except Exception as e:
                    logger.error(e)

            for func in self.item_funcs:
                if func["sub"]:
                    subs = [sub for sub in self.subreddits if str(sub).lower() == func["sub"].lower()]
                else:
                    subs = self.subreddits
                for sub in subs:
                    sub_func = getattr(sub, func["sub_func"])
                    try:
                        async for post in sub_func():
                            found = True
                            await func["func"](post)
                    except Exception as e:
                        logger.error(e)

            for func in self.action_funcs:
                if func["sub"]:
                    subs = [sub for sub in self.subreddits if str(sub).lower() == func["sub"].lower()]
                else:
                    subs = self.subreddits
                for sub in subs:
                    try:
                        async for action in sub.get_mod_actions(func["mods"]):
                            found = True
                            await func["func"](action)
                    except Exception as e:
                        logger.error(e)

            if self.bot is not None and self.change_presence:
                try:
                    await self.bot.change_presence(activity=None)
                except Exception as e:
                    logger.error(e)

            if not found:
                wait_time = counter.count()
            else:
                wait_time = counter.reset()

            await asyncio.sleep(wait_time)

    def mod_actions(self, *args, **kwargs):
        def assign(func):
            self.add_mod_actions_func(func, *args, **kwargs)
            return func

        return assign

    def add_mod_actions_func(self, func, *args, **kwargs):
        if asyncio.iscoroutinefunction(func):
            self.action_funcs.append({
                "func": func,
                "mods": kwargs["mods"] if "mods" in kwargs else list(args),
                "sub": kwargs["subreddit"] if "subreddit" in kwargs else None
            })

    async def get_item(self, c):
        s = str(c) if not isinstance(c, discord.Embed) else json.dumps(c.to_dict())
        return await reddit_helper.get_item(self.reddit, self.subreddits, s)

    def get_reactions_embed(self):
        embed = discord.Embed(
            colour=self.embed_color
        )
        embed.title = "Configured reactions"
        for sub in self.subreddits:
            embed.add_field(name="/r/" + str(sub),
                            value="\n".join([str(r) for r in sub.reactions]),
                            inline=False)
        return embed

    def get_subreddits_embed(self):
        embed = discord.Embed(
            colour=self.embed_color
        )
        embed.title = "Subreddits' statuses"
        embed.description = "\n".join([s.get_status() for s in self.subreddits])
        return embed

    def run(self):
        dir_path = os.path.dirname(os.path.realpath(__file__))
        with open(dir_path + "/WELCOME.md") as f:
            print("")
            BOLD = '\033[1m'
            END = '\033[0m'
            print(re.sub(r"\*\*(.+)\*\*", r"{}\1{}".format(BOLD, END), f.read()))
            print("")

        if len(self.item_funcs) > 0 or len(self.action_funcs) > 0:
            self.loop.create_task(self.send_items())
